/*
 *      b16api.c
 *
 *      Copyright (c) 1994 by Borland International
 *      All Rights Reserved.
 *
 */

#include <windows.h>
#pragma hdrstop

#include <stdio.h>
#include <dir.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>
#include <mem.h>

#include "basetype.h"
#include "oldexe.h"
#include "os2.h"
#include "predef.h"

#define F_API_FULLYIMPLEMENTED		0x00000001
#define F_API_PARTIALLYIMPLEMENTED	0x00000002
#define F_API_STUBBED				0x00000004

#define F_IMP_ORDINAL				0x01

typedef struct	{
	
	union	{
		
		bits16	ordinal;
		bits16	nameOffset;
		
	} 	which;

	bits32	references;
	bits8	flags;
	
}	Import;

typedef struct	{
	
	bits16	nImports;
	bits16	nAllocated;
	Import	*imports;
	
}	ImportedModule;

static bits16			*moduleTable;
static char				*importNames;
static bits16			nModules;
static ImportedModule	*modules;

static void addImportByName (bits16 module, bits16 nameOffset)
{
	ImportedModule	*pMod;
	Import			*tmp;
	bits16			i;
	Import			*newImport = NULL;

	pMod = modules + module;
	if	(pMod->nImports == pMod->nAllocated)
	{
		pMod->nAllocated	+= 10;
		pMod->imports	 	 = (Import *)realloc (pMod->imports, pMod->nAllocated * sizeof (Import));
		assert (pMod->imports);
	}

	tmp = pMod->imports;
	for	(i = 0; i < pMod->nImports; i++, tmp++)
	{
		if	(tmp->flags == 0 && tmp->which.nameOffset == nameOffset)
		{
			newImport = tmp;
			newImport->references++;
			break;
		}
	}

	if	(!newImport)
	{
		newImport = pMod->imports + pMod->nImports++;

		newImport->flags			= 0;
		newImport->which.nameOffset	= nameOffset;
		newImport->references		= 1;
	}
}

static void addImportByOrdinal (bits16 module, bits16 ordinal)
{
	ImportedModule	*pMod;
	Import			*tmp;
	bits16			i;
	Import			*newImport = NULL;

	pMod = modules + module;
	if	(pMod->nImports == pMod->nAllocated)
	{
		pMod->nAllocated	+= 10;
		pMod->imports	 	 = realloc (pMod->imports, pMod->nAllocated * sizeof (Import));
		assert (pMod->imports);
	}
	
	tmp = pMod->imports;
	for	(i = 0; i < pMod->nImports; i++, tmp++)
	{
		if	((tmp->flags & F_IMP_ORDINAL) && tmp->which.ordinal == ordinal)
		{
			newImport = tmp;
			newImport->references++;
			break;
		}
	}

	if	(!newImport)
	{
		newImport = pMod->imports + pMod->nImports++;

		newImport->flags			= F_IMP_ORDINAL;
		newImport->which.ordinal	= ordinal;
		newImport->references		= 1;
	}
}

static void usage (void)
{
	printf ("\nB16API [-s] <filespec>[.DLL]\n"
			"  Prints compatibility report of 16bit EXE or DLL\n");
}

static void setExtension (char *dest, const char *path, const char *newExt, int force)
{
    char	drive	[MAXDRIVE];
    char	direct	[MAXDIR];
    char	name	[MAXFILE];
    char	ext		[MAXEXT];

	fnsplit (path, drive, direct, name, ext);
	if	(force || !*ext)
		fnmerge (dest, drive, direct, name, newExt);
	else
		fnmerge (dest, drive, direct, name, ext);
}

static boolean compareModules (char *name, char **modules)
{
	while	(*modules)
	{
		if	(!stricmp (name, *modules++))
			return TRUE;
	}
	return FALSE;
}

static char *systemModules [] =
{
	"KERNEL",
	"USER",
	"KEYBOARD",
	"GDI",
	"WIN87EM",
	0
};

static boolean littlefootModule (char *name)
{
	static char *littlefootModules [] =
	{
		"KERNEL",
		"USER",
		"KEYBOARD",
		"RTM",
		0
	};

	return compareModules (name, littlefootModules);
}

extern KnownOrdinal	kernelOrdinals;
extern KnownOrdinal	userOrdinals;
extern KnownOrdinal	gdiOrdinals;
extern KnownOrdinal	keyboardOrdinals;
extern KnownOrdinal	win87emOrdinals;

static KnownOrdinal *findPredefinitions (const char *modName)
{
	if	(!stricmp (modName, "KERNEL"))
		return &kernelOrdinals;
	
	if	(!stricmp (modName, "USER"))
		return &userOrdinals;
	
	if	(!stricmp (modName, "GDI"))
		return &gdiOrdinals;
	
	if	(!stricmp (modName, "KEYBOARD"))
		return &keyboardOrdinals;
	
	if	(!stricmp (modName, "WIN87EM"))
		return &win87emOrdinals;

	return NULL;
}

static const char *findName (KnownOrdinal *list, bits16 ordinal)
{
	do	{

		if	(list->ordinal == ordinal)
			return list->name;
		
		list++;
		
	}	while (list->ordinal != 0);

	return NULL;
}

static void reportSupported (void)
{
	HMODULE			hMod;
	char			**modNames;
	KnownOrdinal	*list;

	printf ("\nSupported function summary:\n");

	modNames = systemModules;
	while	(*modNames)
	{
		hMod = GetModuleHandle (*modNames);
		if	(hMod)
		{
			printf ("\nFunctions supported for module %s:\n", *modNames);
			list = findPredefinitions (*modNames);
			assert (list);
			while	(list->ordinal)
			{
				if	(GetProcAddress (hMod, (LPCSTR)list->ordinal))
					printf ("%s\n", list->name);
				list++;
			}
		}
		modNames++;
	}
}

static void reportImports (void)
{
	bits16			i;
	bits16			j;
	ImportedModule	*pMod;
	Import			*tmp;
	char			modName [9];
	KnownOrdinal	*predefs;

	pMod = modules;

	printf ("\nModule         Implemented by RTM\n");
	printf (  "---------------------------------\n");
	
	for	(i = 0; i < nModules; i++, pMod++)
	{
		tmp = pMod->imports;
		if	(tmp)
		{
			char	*pName;

			pName = importNames + moduleTable [i];
			strncpy (modName, pName + 1, *pName);
			*(modName + *pName) = 0;
			printf ("%-15s%s\n",
					modName,
					littlefootModule (modName) ? "Yes" : "No");
		}
	}
			
	pMod = modules;

	for	(i = 0; i < nModules; i++, pMod++)
	{
		tmp = pMod->imports;
		if	(tmp)
		{
			boolean		isLittlefoot;
			HMODULE		hMod;
			char		*pName;

			pName = importNames + moduleTable [i];
			strncpy (modName, pName + 1, *pName);
			*(modName + *pName) = 0;
			printf ("\nImports from module %s:\n", modName);
			
			printf ("\n#References  API                                     Implemented by RTM\n");
			printf ("-------------------------------------------------------------------------------\n");
	
			predefs = findPredefinitions (modName);
			isLittlefoot = littlefootModule (modName);

			if	(isLittlefoot)
			{
				hMod = GetModuleHandle (modName);
				assert (hMod);
			}
			
			for	(j = 0; j < pMod->nImports; j++, tmp++)
			{
				printf ("%08ld     ", tmp->references);
				
				if	(tmp->flags & F_IMP_ORDINAL)
				{
					const char	*entryName;

					if	(predefs)
					{
						entryName = findName (predefs, tmp->which.ordinal);
						if	(entryName)
							printf ("%-40s", entryName);
						else
							printf ("%-40d", tmp->which.ordinal);

					}
					else
						printf ("%-40d", tmp->which.ordinal);
					
					if	(isLittlefoot)
					{
						if	(GetProcAddress (hMod, (LPCSTR)tmp->which.ordinal))
							printf ("Yes\n");
						else
							printf ("No\n");
					}
					else
						printf ("No\n");
				}
				else
				{
					printf ("%-40s", importNames + tmp->which.nameOffset);
					
					if	(isLittlefoot)
					{
						if	(GetProcAddress (hMod, (LPCSTR)tmp->which.nameOffset))
							printf ("Yes\n");
						else
							printf ("No\n");
					}
					else
						printf ("No\n");
				}
			}
		}
	}
}

static void loadTables (FILE *fp, TNewExeHeader *hdr, bits32 base)
{
	bits16	size;
	
	/* Load the module table */
	size = (bits16)(hdr->ne_imptab - hdr->ne_modtab);
	
	if	( size == 0 )
		moduleTable = NULL;
	else
	{
		moduleTable = (bits16 *)malloc (size + 2);
		assert (moduleTable);

		nModules = (size + 2) / 2;
		modules = (ImportedModule *)malloc (nModules * sizeof (ImportedModule));
		assert (modules);
		memset (modules, 0, nModules * sizeof (ImportedModule));
		
		fseek (fp, base + (long)hdr->ne_modtab, SEEK_SET);
		fread (moduleTable + 1, size, 1, fp);
	}

	/* Load the Imported Names table */
	size = (bits16)(hdr->ne_enttab - hdr->ne_modtab);
	
	if	(size == 0)
		importNames = NULL;
	else
	{
		importNames = malloc (size);
		assert (importNames);
		
		fseek (fp, base + (long)hdr->ne_imptab, SEEK_SET);
		fread (importNames, size, 1, fp);
	}
}

static void cacheImports (FILE *fp, TNewExeHeader *hdr, bits32 base)
{
	SEGMENT_RECORD		srec;
	RELOCATION_RECORD	rRec;
	bits16				i;
	bits16				j;
	bits16				count;
	bits32				cloc;
	
	fseek (fp, base + (long)hdr->ne_segtab, SEEK_SET);
	for	(i=1 ; i <= hdr->ne_cseg ; ++i)
	{
		fread (&srec, sizeof (SEGMENT_RECORD), 1, fp);

		if	(srec.reloc_info)
		{
			cloc = ftell (fp);
			fseek (fp, ((long)srec.sector_offset << hdr->ne_align) + (long)srec.segment_length, SEEK_SET);
			fread (&count, sizeof (count), 1, fp);

			for	(j = 0; j < count; j++)
			{
				fread (&rRec, sizeof (RELOCATION_RECORD), 1, fp);

				switch	(rRec.entryType & 3)
				{
				case	1:
					addImportByOrdinal (rRec.i.o.index, rRec.i.o.ordinal);
					break;
												
				case	2:
					addImportByName (rRec.i.o.index, rRec.i.n.nameOff);
					break;
				}
			}
			
			fseek (fp, cloc, SEEK_SET);
		}
	}
}

static int report (FILE *fp)
{
	OldExeHdr		oldHdr;
	TNewExeHeader	hdr;
	
	fread (&oldHdr, sizeof (oldHdr), 1, fp);
	if	(oldHdr.signature != MZSIGNATURE	||
		 oldHdr.reloc_offset != 0x40)
	{
		printf ("Not an NE file\n");
		return 1;
	}

	fseek (fp, oldHdr.NE_LE_start, SEEK_SET);
	fread (&hdr, sizeof (hdr), 1, fp);
	if	(hdr.ne_magic != NEMAGIC)
	{
		printf ("Not an NE file\n");
		return 1;
	}

	loadTables (fp, &hdr, oldHdr.NE_LE_start);
	cacheImports (fp, &hdr, oldHdr.NE_LE_start);
	reportImports ();

	return 0;
}

static void banner (void)
{
	printf ("Borland 16bit API compatibility tester\n"
			"Version .01  Copyright (c) 1994 Borland International\n");
}

int main (int argc, char **argv)
{
	char		path [MAXPATH];
	FILE		*fp;
	int			res;

	banner ();
	
	if	(argc < 2)
	{
		usage ();
		return 0;
	}

	if	(*argv [1] == '-' || *argv [1] == '/')
	{
		char	*tmp;

		tmp = argv [1] + 1;
		if	(*tmp == 's')
		{
			reportSupported ();
			return 0;
		}
	}

	setExtension (path, argv [1], ".dll", 0);
	fp = fopen (path, "rb");
	if	(!fp)
	{
		printf ("Couldn't open '%s'\n", path);
		return 2;
	}

	res = report (fp);
	fclose (fp);

	return res;
}

/*
**  End.
*/

